
// --------------------------------------------------------------------------------------

//		Sprite work by The Angry, with quality input by the SRB2 Brasil community
//		Dirfex coding by Jon, with aditional work by Xelork on the drip meme pose

//		This little guy took over two years to come out because of various roadblocks,
//		but we're glad to finally be releasing him! Enjoy!

//		Notorious testers:	- Phatic Static
//							- Nico the Moon
//							- TheLastKitchenGun
//							- JosephLol

// --------------------------------------------------------------------------------------


//Variables for ease of tweaking

local MAXSTAMINA = 100				//Maximum running stamina
local STAMDAMAGE_ACTIVE = 1			//Stamina cost for regular dashing
local STAMDAMAGE_JUMP = 19			//Stamina cost for a dash jump

local STOMPRANGE = 150				//Stomp impact range for damaging enemies
local STOMP_EMERGENCY = 4			//How many seconds before a stomp is automatically cancelled

local DASHCOYOTE = 7				//Frames of coyote time for the dash jump

local BRIDGELIMIT = 20				//How many bridge segments dirfex can create at once

local SPIKECOOLDOWN = 5				//How long to wait between each spike circle's creation
local SPIKEDIST = 39				//How far away from dirfex should his spikes spawn
local DIRSPIKECIRCLES = 3			//How many circles of spikes dirfex should create per stomp
local DIRSPIKEPERCIRCLE = 8			//How many spikes a spike circle should have

local RINGREQMIN = 10				//Minimum amount of rings required for an ice shield
local RINGREQADD = 20				//How many rings should be added to the requirement for the ice barrier whenever it is summoned
local BARRIERTICS = 10				//How many seconds to wait before the ice barrier can regenerate

//Instead of checking the same stuff over and over just use this function
local function validDirf(mo)
	if (not mo or not mo.valid) or
	   (not mo.player or not mo.player.valid) or
	   (not mo.player.dfx) or
	   (mo.skin ~= "dirfex") or
	   (maptol & TOL_NIGHTS)
		return false
	end
	return true
end

//Rawset function to check if dirfex can create an ice bridge
rawset(_G, "CanMakeBridge", function(mo)
	if not validDirf(mo) return false end

	if abs(mo.watertop - mo.z) < 4*FRACUNIT //mo.watertop > mo.z - 1000*FRACUNIT
	and mo.state == S_PLAY_DIRDASH
	and mo.floorrover
	and not (mo.floorrover.flags & FF_GOOWATER)
	and not P_ThingOnSpecial3DFloor(mo) //I hope this doesn't bite me in the ass later
	and not mo.player.dfx.cantbridge
	and not (mo.eflags & MFE_UNDERWATER)
		return true
	end
	return false
end)

//Let's get ready for our spin stomp
local function prepSpinStomp(p)
	if not validDirf(p.mo) return false end
	S_StartSound(p.mo, sfx_spndsh)
	p.mo.state = S_PLAY_DIRSPINSTOMP
	p.pflags = $|PF_THOKKED & ~PF_SPINNING
	p.dfx.stompMom = 10*FRACUNIT
	p.dfx.stompEmergency = STOMP_EMERGENCY*TICRATE
	p.dfx.squashxscale = p.mo.spritexscale
	p.dfx.squashyscale = p.mo.spriteyscale
	p.dfx.correctSquash = true
end

//Function to reset anything dirfex related
local function dirReset(p, keepringshield)
	if not p.valid return end
	if not p.mo.valid return end
	if p.dfx
		if p.dfx.carrygoyle and p.dfx.carrygoyle.valid then P_RemoveMobj(p.dfx.carrygoyle) end
		p.dfx = nil
	end
	p.mo.rollangle = 0
	p.mo.spritexscale = FRACUNIT
	p.mo.spriteyscale = FRACUNIT
	p.mo.shadowscale = FRACUNIT
	if not keepringshield then p.dirringshi = nil end
end

//For when we crash into stuff. Made into a function instead of MobjCollide so we can toss bosses around
local function dashBonk(pmo, blocker, line)
	if not validDirf(pmo) return end
	if blocker
		if not blocker.valid return end
		//monitor interaction
		if blocker.flags & MF_MONITOR 
		and (pmo.state == S_PLAY_DIRDASH or pmo.state == S_PLAY_DIRJUDASH) and pmo.player.speed >= pmo.player.normalspeed
			P_DamageMobj(blocker, pmo, pmo) 
			return
		end
		//pushable/gargoyle interactions
		if (blocker.flags & MF_PUSHABLE) or (blocker.flags & MF_BOSS) or (blocker.flags & MF_ENEMY and blocker.health > 1)
		and (pmo.state == S_PLAY_DIRDASH or pmo.state == S_PLAY_DIRJUDASH) and pmo.player.speed >= pmo.player.normalspeed
			P_InstaThrust(blocker, pmo.angle, pmo.player.speed/2)
			if blocker.type == mobjinfo[MT_GARGOYLE].doomednum then S_StartSound(blocker, sfx_statu2) end
		end
	end
	
	//Get bonked and thrown back slightly
	if (pmo.state == S_PLAY_DIRDASH or pmo.state == S_PLAY_DIRJUDASH) and pmo.player.speed >= pmo.player.normalspeed
		pmo.momx = -$/8
		pmo.momy = -$/8
		P_SetObjectMomZ(pmo, 8*FRACUNIT)
		pmo.state = S_PLAY_FALL
		S_StartSound(pmo, sfx_s3k49)
		for i = 1, 8 do
			local zang = i*ANGLE_45
			local whatToSpawn = nil
			if pmo.eflags & MFE_UNDERWATER then whatToSpawn = MT_MEDIUMBUBBLE else whatToSpawn = MT_SPINDUST end
			local dust = P_SpawnMobj(pmo.x, pmo.y, pmo.z + (pmo.height/4)*3, whatToSpawn)
			dust.angle = pmo.angle
			dust.flags = $|MF_NOGRAVITY
			P_InstaThrust(dust, dust.angle+ANGLE_90, FixedMul(cos(zang), FRACUNIT*12))
			P_SetObjectMomZ(dust, FixedMul(sin(zang), FRACUNIT*3), false)
		end
	end
end

//Function used to yeet enemies away from dirfex during his dash or stomp
local function thingYeeter(enemy, toucher, ignoreState)
	if not enemy.valid return end
	if not validDirf(toucher) return end

	if not (enemy.flags & MF_ENEMY) 
	and not (enemy.flags & MF_BOSS)
	and not (enemy.flags & MF_MONITOR) then return end
	
	if ( (toucher.state == S_PLAY_DIRDASH or toucher.state == S_PLAY_DIRJUDASH) and toucher.player.speed > toucher.player.normalspeed - 6*FRACUNIT )
	or ignoreState
		//Only do the stuff down here if it's a grounded enemy. Flying ones just get damaged.
		if (not (enemy.flags & MF_NOGRAVITY)) and enemy.health == 1
			enemy.knockAng = R_PointToAngle2(toucher.x, toucher.y, enemy.x, enemy.y)
			enemy.flags = $ & ~MF_SPECIAL
			P_InstaThrust(enemy, enemy.knockAng, toucher.player.speed + (toucher.player.speed/5))
			enemy.z = $+FRACUNIT*3
			P_SetObjectMomZ(enemy, 5*FRACUNIT)
			enemy.ohgodohfuck = true
			enemy.killer = toucher
			enemy.colortimer = 2
			enemy.origcolor = enemy.color
			enemy.colorized = true
			enemy.color = SKINCOLOR_SALMON
			S_StartSound(enemy, sfx_s3k6e)
		elseif (enemy.flags & MF_NOGRAVITY)
			P_DamageMobj(enemy, toucher, toucher)
		elseif (not (enemy.flags & MF_NOGRAVITY)) and enemy.health > 1
			dashBonk(toucher, enemy, nil)
			P_DamageMobj(enemy, toucher, toucher)
		end
		if (enemy.flags & MF_BOSS) then dashBonk(toucher, enemy, nil) end
		return true
	end
end

local function spikeSpawner(mo, spikenum, scale, angle, radius, spx, spy, spz)
	if not validDirf(mo) return end
	if not scale then scale = FRACUNIT end
	if not angle then angle = ANGLE_90 end
	local spikeSound = false
	local spikeheight = 1
	if mo.player.dfx.spikesSpawned then spikeheight = mo.player.dfx.spikesSpawned end
	if mo.player.powers[pw_super] then radius = (SPIKEDIST/10) * 14 * spikeheight end
	if (mo.eflags & MFE_VERTICALFLIP)
		spz = $ + mo.height
		if spikeheight > 1 then spz = $ + 29 * FRACUNIT end // I hate reverse gravity so god damn much
	end
	radius = FixedMul($, mo.scale)

	for i = 1, spikenum
		local zang = i * ( FixedAngle( (360/spikenum)*FRACUNIT ) )
		local spike = P_SpawnMobj(spx, spy, spz, MT_DICESPIKE)

		spike.angle = mo.angle + zang
		spike.height = 24 * FRACUNIT * spikeheight
		spike.spritexscale = FixedMul($ - $/4, mo.scale)
		spike.spriteyscale = 1
		spike.target = mo
		spike.fuse = 12
		spike.destyscale = FixedMul(scale, mo.scale)
		P_TeleportMove(spike, spike.x + radius*cos(spike.angle), spike.y + radius*sin(spike.angle), spike.z)
		if (mo.eflags & MFE_VERTICALFLIP) 
			spike.z = spike.ceilingz
			spike.eflags = $|MFE_VERTICALFLIP
		else 
			spike.z = spike.floorz
		end
		spike.angle = $ + ANGLE_180
		if abs(spz - spike.z) > 25*FRACUNIT or not P_IsObjectOnGround(spike) then P_RemoveMobj(spike) end
		if spike.valid and not spikeSound 
			S_StartSound(spike, sfx_dbri)
			spikeSound = true
		end
	end
end

//Also do that when we die
addHook("MobjDeath", function(pmo, inflictor, source, dtype)
	if not pmo.valid return end
	if not pmo.player.valid return end
	if pmo.skin == "dirfex" then dirReset(pmo.player) end
end, MT_PLAYER)

//And when we respawn
addHook("PlayerSpawn", function(p)
	if not p.mo return end
	if not p.mo.valid return end
	if not p.mo.player.valid return end
	if p.mo.skin == "dirfex" then dirReset(p.mo.player) p.dirringshi = RINGREQMIN p.powers[pw_underwater] = 60*TICRATE end
end)

//Function to prepare dirfex's dash. All dash initialization stats go here.
local function dirPrepDash(p)
	if not p.valid return end
	if not p.mo.valid return end
	p.dfx.startSpeed = p.speed					//At what speed did we start our dash?
	p.dfx.buildSpeed = 0						//Set our buildup speed (the speed we'll ease into top speed) to zero
	if p.dfx.startSpeed < p.normalspeed/3
		p.dfx.spBuildRate = FRACUNIT/18			//At low speed, build up to your dash slowly
	elseif p.dfx.startSpeed < ((p.normalspeed/3) * 2)
		p.dfx.spBuildRate = FRACUNIT/12			//At mid speed, build up faster
	else
		p.dfx.spBuildRate = FRACUNIT/6			//At near top speed, build up almost instantly
	end
end

//Function to prepare dash leaning, both grounded and in the air
local function dirPrepLean(p)
	if not p.valid return end
	if not p.mo.valid return end
	if not p.dfx.lastAngle then p.dfx.lastAngle = 0 end
	local angdiff = p.dfx.lastAngle - p.mo.angle
	local LR = 0
	if angdiff > 0 then LR = -1 else LR = 1 end
	if p.dfx.rollReturn < 2 or LR == p.dfx.LR
		if abs(angdiff) > FRACUNIT*1000 then p.dfx.rollReturn = 8
		elseif abs(angdiff) > FRACUNIT*500 and p.dfx.rollReturn < 4 then p.dfx.rollReturn = 3 end
		p.dfx.LR = LR
	end
end

//PreThink so we can edit dirfex's inputs if needed
addHook("PreThinkFrame", function()
	for p in players.iterate
		if not validDirf(p.mo) continue end

		//Limit our sideways movement during the grounded dash
		if p.mo.state == S_PLAY_DIRDASH then p.cmd.sidemove = 0 end

		//Preventing dirfex from going nuclear when holding SPIN after falling off a platform while dashing
		if p.dfx.dashLiftoff then p.cmd.buttons = $ & ~BT_SPIN end

		//Limit our direct control over dirfex ever so slightly during a stomp
		if p.mo.state == S_PLAY_DIRSPINSTOMP
			local stompbrake = p.powers[pw_super] and 1 or 50
			p.cmd.forwardmove = $/stompbrake
			p.cmd.sidemove = $/stompbrake
		end

		if (p.pflags & PF_ANALOGMODE) and p.mo.state == S_PLAY_DIRDASH
			p.cmd.forwardmove = 50
		end

	end
end)

//PlayerThink for most of our stuff
addHook("PlayerThink", function(p)
	if not p.valid or not p.mo or not p.mo.valid return end
	if maptol & TOL_NIGHTS return end

	if p.dorflex == nil then p.dorflex = false end

	if p.mo.skin ~= "dirfex"
		if p.dfx then dirReset(p, true) end //We aren't dirfex, but still have his stats. Reset almost everything.
		return 
	end
	if p.playerstate ~= PST_LIVE return end

	//We became dirfex! if this table exists while using a different skin then reset all our dirfex related variables.
	//This is also used to define our starting variables
	if not p.dfx
		p.dfx = {}
		p.dfx.rollreturn = 0
		p.dfx.lr = 0
		p.dfx.stamina = MAXSTAMINA
		p.dfx.stamhud = 0
	end

	//Making sure we don't slip around on ice
	p.mo.friction = (29 * (FRACUNIT/32)) - FRACUNIT/128

	//If we bump into someone in ringslinger make sure we don't just absorb all their rings for a few tics
	if p.dfx.nofling then p.dfx.nofling = $-1 end

	//Initializing dirfex's ice shield charged by rings. This isn't reset by dirReset unless a new map is loaded.
	if p.dirringshi == nil then p.dirringshi = RINGREQMIN end
	
	//Count down our barrier's timer after it's been shattered. We can only make a new one after this time.
	if p.dfx.barriercooldown
		p.dfx.barriercooldown = $-1
	end

	//Ice sub-shield enabled by collecting rings. Increases ring requirement by 20 every time it's activated
	if not p.powers[pw_super] and not p.dfx.barriercooldown
	and p.dirringshi and p.rings >= p.dirringshi and not p.dfx.icebarrier
		S_StartSound(p.mo, sfx_dbri)
		p.dirringshi = $ + RINGREQADD
		p.dfx.icebarrier = true
		if not p.dfx.iceshield
			p.dfx.iceshield = {}
			for i = 1, 8
				local ang = i*ANGLE_45
				local shi = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z + p.mo.z, MT_DICESHIELD)
				if not (i%2) then shi.updown = 1 else shi.updown = -1 end
				shi.angle = p.mo.angle + ang
				shi.scale = 1
				shi.destscale = FRACUNIT/3
				shi.shadowscale = 0
				shi.dad = p.mo
				p.dfx.iceshield[i] = shi
			end
		end
	end

	//Ignoring Mystic Realm momentum tweaks so dirfex doesn't achieve warp speed instantly in MRCE maps.
	//Also adds hyper form compatibility, albeit without any custom changes.
	if p.mrce 
		p.mrce.customskin = 2
		p.mrce.canhyper = true
		p.mrce.hyperimages = true
		p.mrce.ultrastar = true
	end

	//Player carrying! Let's allow players to be carried by dirfex as long as he's in an idle state and the player isn't flying
	if (p.panim == PA_IDLE or p.mo.state == S_PLAY_DIRCARRY) and not (leveltime%4)
		
		local found = false
		searchBlockmap("objects", function(pmo, mo)
			if not mo.player or not mo.player.valid return end
			if mo.player.panim ~= PA_ABILITY
			and mo.z > pmo.z + pmo.height and mo.z < pmo.z + pmo.height + 55*FRACUNIT
				found = true
				p.mo.state = S_PLAY_DIRCARRY
				p.dfx.playercarry = (R_PointToDist2(pmo.x, pmo.y, mo.x, mo.y) < 32*FRACUNIT and (mo.z - pmo.z)/FRACUNIT < 80 and P_IsObjectOnGround(mo))
			end
		end, p.mo, p.mo.x, p.mo.x + 25*FRACUNIT, p.mo.y, p.mo.y + 25*FRACUNIT)
		
		if p.mo.state == S_PLAY_DIRCARRY and not found
			p.mo.state = S_PLAY_STND
		end

	end

	//If we entered our carrying position that means we can now carry players, so let's spawn something on top of us to hold 'em
	if p.mo.state == S_PLAY_DIRCARRY

		p.mo.sprite2 = p.dfx.playercarry and SPR2_DSEG or SPR2_STND
		//We'll spawn an invisible gargoyle on top of dirfex to be used as a carrying platform
		if not p.dfx.carrygoyle or (p.dfx.carrygoyle and not p.dfx.carrygoyle.valid)
			if p.dfx.carrygoyle and not p.dfx.carrygoyle.valid then p.dfx.carrygoyle = nil end
			p.dfx.carrygoyle = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_GARGOYLE)
			p.dfx.carrygoyle.flags = $ | MF_NOGRAVITY | MF_NOCLIPHEIGHT & ~MF_PUSHABLE & ~MF_SLIDEME
			p.dfx.carrygoyle.flags2 = $ | MF2_DONTDRAW
			p.dfx.carrygoyle.radius = 8*FRACUNIT
		end

		//Put the carrygoyle where it should be
		P_TeleportMove(p.dfx.carrygoyle, p.mo.x, p.mo.y, p.mo.z + p.mo.height/2)
		
		if p.cmd.forwardmove or p.cmd.sidemove or (p.cmd.buttons & BT_JUMP) or (p.cmd.buttons & BT_SPIN)
			if p.dfx.carrygoyle and p.dfx.carrygoyle.valid then P_RemoveMobj(p.dfx.carrygoyle) end
			p.dfx.carrygoyle = nil
			p.mo.state = S_PLAY_WALK
		end
	else
		if p.dfx.carrygoyle
			if p.dfx.carrygoyle.valid then P_RemoveMobj(p.dfx.carrygoyle) end
			p.dfx.carrygoyle = nil
		end
	end

	//Let's make an ice bridge. First of all check our bridge making requirements and whether or not we have -
	//- a "leading bridge piece" created or not. If not, make a leading piece!
	//We will also run this if we were already creating a bridge before, but did a dash jump while doing so
	if CanMakeBridge(p.mo) and ((not p.dfx.bridge) or p.dfx.bridgejump)
		p.dfx.bridge = P_SpawnMobjFromMobj(p.mo, 0, 0, -10*FRACUNIT, MT_DBRIPART)
		if p.dfx.bridgejump and p.dfx.bridgejump.valid
			p.dfx.bridgejump.first = false
			p.dfx.bridgejump.afterme = p.dfx.bridge
		else
			p.dfx.firstbridge = p.dfx.bridge
		end
		p.dfx.bridge.fuse = 5*TICRATE
		p.dfx.bridge.scale = FRACUNIT + FRACUNIT/2
		p.dfx.bridge.spritexscale = 1
		p.dfx.bridge.spriteyscale = 1
		p.dfx.bridge.dad = p.mo
		p.dfx.bridge.first = true
		p.dfx.bridge.renderflags = $|RF_NOSPLATBILLBOARD
		if p.dfx.bridgecount ~= nil then p.dfx.bridgecount = $+1 else p.dfx.bridgecount = 1 end
		S_StartSound(p.mo, sfx_dbri)
	end

	//We're might be going over our max bridge limit, so let's handle this in a FIFO manner
	if p.dfx.firstbridge and p.dfx.firstbridge.valid
		if p.dfx.bridgecount > BRIDGELIMIT
			p.dfx.firstbridge.removed = true
			p.dfx.firstbridge.fuse = 1
			p.dfx.firstbridge = p.dfx.firstbridge.afterme
			p.dfx.bridgecount = $-1
		end
	end

	//Plenty of bridge sanity checks. This needs some cleaning up...
	if p.dfx.bridgejump and P_IsObjectOnGround(p.mo) then p.dfx.bridgejump = nil end
	if p.dfx.bridge and not p.dfx.bridge.valid then p.dfx.bridge = nil end
	if p.mo.eflags & MFE_UNDERWATER or abs(p.mo.watertop - p.mo.z) > 8*FRACUNIT then p.dfx.cantbridge = true end
	if (p.mo.watertop == p.mo.z - 1000*FRACUNIT) or abs(p.mo.watertop - p.mo.z) <= 8*FRACUNIT then p.dfx.cantbridge = false end

	//Var initialization
	if p.dfx.rollReturn == nil then p.dfx.rollReturn = 0 end
	if p.dfx.LR == nil then p.dfx.LR = 0 end
	if p.dfx.stamina == nil then p.dfx.stamina = MAXSTAMINA end
	if p.dfx.stamHud == nil then p.dfx.stamHud = 0 end

	//Extra breathing time underwater and in space
	if not (p.mo.eflags & MFE_UNDERWATER) then p.powers[pw_underwater] = 60*TICRATE end
	if not P_InSpaceSector(p.mo) then p.powers[pw_spacetime] = 24*TICRATE end

	//HUD stuff for the stamina bar
	if p.dfx.stamina < 100 and p.dfx.stamHud < 8 then p.dfx.stamHud = $+1
	elseif p.dfx.stamina >= 100 and p.dfx.stamHud > 0 then p.dfx.stamHud = $-1 end

	//Setting our dash speed based on where we are and if we have speed shoes
	p.dfx.speed = p.normalspeed + (p.normalspeed/3)
	if (p.powers[pw_sneakers] or p.powers[pw_super]) then p.dfx.speed = (p.normalspeed*2) + (p.normalspeed/4) end
	if p.mo.eflags & MFE_UNDERWATER
		if (p.powers[pw_sneakers] or p.powers[pw_super]) then p.dfx.speed = p.normalspeed 
		else p.dfx.speed = (p.normalspeed/2) + p.normalspeed/5 end
	end

	//Coyote time for falling off a platform while dashing
	if p.dashcoyote and not P_IsObjectOnGround(p.mo) 
		p.dashcoyote = $-1
	end

	//Start our dash when pressing C1 when grounded
	if (p.cmd.buttons & BT_CUSTOM1) and not p.exiting// and not p.dfx.stasisTime
		if P_IsObjectOnGround(p.mo) 
		and not p.dfx.c1Press 
		and not (p.pflags & PF_SPINNING)
		and p.mo.state ~= S_DRIP
		and (p.dfx.stamina > 0 and not p.dfx.noStam)
			p.mo.state = S_PLAY_DIRDASH
			dirPrepDash(p)
		end
		p.dfx.c1Press = true
	else
		p.dfx.c1Press = false
	end

	//Stomp time, let's set our momz then do a grounded shockwave when we land (handled in PostThink)
	if p.mo.state == S_PLAY_DIRSPINSTOMP
		if p.dfx.stompMom ~= nil
			if not (leveltime%3) P_SpawnGhostMobj(p.mo) end
			P_SetObjectMomZ(p.mo, p.dfx.stompMom, false)
			if p.dfx.stompMom > -38*FRACUNIT then p.dfx.stompMom = $ - (FRACUNIT + (FRACUNIT/2)) end
			if not p.powers[pw_super]
				P_InstaThrust(p.mo, p.drawangle, p.speed - (p.speed/7) )
			end
			if p.dfx.stompMom < 0 and p.mo.spritexscale > FRACUNIT - (FRACUNIT/4)
				p.mo.spritexscale = $ - FRACUNIT/40
				p.mo.spriteyscale = $ + FRACUNIT/40
			end
		end
		if p.dfx.stompEmergency and p.dfx.stompEmergency > 0 then p.dfx.stompEmergency = $-1 end
		if P_IsObjectInGoop(p.mo) or not p.dfx.stompEmergency
			p.dfx.stompMom = nil
			p.mo.state = S_PLAY_JUMP
		end
	end

	//We're dashing. Let's do some stuff.
	if p.mo.state == S_PLAY_DIRDASH

		p.pflags = $ & ~PF_SPINNING

		//Preventing silly moonwalk shenanigans
		if p.mo.state == S_PLAY_DIRDASH
			p.drawangle = p.mo.angle
		end

		if p.dfx.c1Press and p.dfx.stamina > 0
			if p.cmd.forwardmove == 0 then p.cmd.forwardmove = 1 end //for some reason we need forwardmove to make momentum count. odd.

			//Grounded stuff
			if P_IsObjectOnGround(p.mo) 
				//We're on the ground, so let's make some noise, spawn some objects and move our legs faster
				if p.speed > FixedMul(p.normalspeed, p.mo.scale)
					local stepSound = sfx_s23b
					if p.dfx.stamina < 30 then stepSound = sfx_dtir end
					if not (leveltime%2) and not (p.powers[pw_sneakers] or p.powers[pw_super]) then p.dfx.stamina = $ - STAMDAMAGE_ACTIVE end
					if not (leveltime%4) P_SpawnSkidDust(p, 20*FRACUNIT) end
					if not (leveltime%8) S_StopSoundByID(p.mo, stepSound) S_StartSound(p.mo, stepSound) end
					//Dust stuff
					local dustangle = ANGLE_67h
					local dustscale = 2
					local dustdestscale = FRACUNIT+(FRACUNIT/2)
					//Create a shield
					if not p.dfx.dashShield
						p.dfx.dashShield = P_SpawnMobjFromMobj(p.mo, 0, 0, 0, MT_DSHI)
						p.dfx.dashShield.scale = (FRACUNIT/20) + 5
						p.dfx.dashShield.scalespeed = FRACUNIT/10
						p.dfx.dashShield.destscale = FRACUNIT
						p.dfx.dashShield.dad = p.mo
						if P_MobjFlip(p.mo) > 0 then p.dfx.dashShield.spriteyoffset = $ + p.mo.height/2 
						else p.dfx.dashShield.spriteyoffset = $ - p.mo.height/4 end
					end
				end
				if p.mo.eflags & MFE_UNDERWATER 
					P_SpawnSkidDust(p, 15*FRACUNIT)
					if not (leveltime%10) S_StopSoundByID(p.mo, sfx_splash) S_StartSound(p.mo, sfx_splash) end
				end 

				//Also give us some coyote time so we can jump in the air mid-dash for 4 tics
				p.dashcoyote = DASHCOYOTE

				//And finally let's give us some slight rollangle stuff for that nice game juice
				if not(p.mo.eflags & MFE_UNDERWATER) then dirPrepLean(p) end
				//Now that we set our DirRollReturn we can use this to tweak our rollangle outside of this all encompassing IF
				

			//We're in the air! This is used to ensure we can cantinue running after falling back to the ground.
			else
				p.dfx.dashLiftoff = true
			end
			//Let's build up some speed through InSine easing
			local newmom = ease.insine(p.dfx.buildSpeed, p.dfx.startSpeed, p.dfx.speed)
			newmom = FixedMul($, p.mo.scale)
			local whereToDash = p.mo.angle
			if P_IsObjectOnGround(p.mo) and not (p.mo.eflags & MFE_UNDERWATER)
				if p.speed < p.dfx.speed - 3*FRACUNIT
					P_InstaThrust(p.mo, whereToDash, newmom)
				elseif p.speed > p.dfx.speed - 3*FRACUNIT and p.mo.rollangle
					P_InstaThrust(p.mo, whereToDash, FixedMul(p.speed, p.mo.scale))
				else
					P_Thrust(p.mo, whereToDash, FixedMul(4*FRACUNIT + (FRACUNIT/6), p.mo.scale))
				end
			elseif p.mo.eflags & MFE_UNDERWATER
				P_InstaThrust(p.mo, whereToDash, newmom)
			end
			if p.dfx.buildSpeed < FRACUNIT then p.dfx.buildSpeed = $+p.dfx.spBuildRate end
		else
			//We let go off of C1 or are out of stamina, so let's return to normal
			if P_IsObjectOnGround(p.mo) then p.mo.state = S_PLAY_WALK
			else 
				p.mo.state = S_PLAY_FALL
				p.pflags = $ & ~PF_SPINNING
			end
		end
	end

	//Things to do in any dash case, such as spawning snowflakes
	if p.mo.state == S_PLAY_DIRDASH or p.mo.state == S_PLAY_DIRJUDASH

		//If we're dashing we can run on water!
		p.charflags = $|SF_RUNONWATER

		//Effect spawning
		if p.speed > p.normalspeed 

			//Snowflakes
			if not (leveltime%3)
				local thicc = p.mo.radius/FRACUNIT
				local tall = p.mo.height/FRACUNIT
				local flake = P_SpawnMobjFromMobj(p.mo, P_RandomRange(-thicc, thicc)*FRACUNIT, P_RandomRange(-thicc, thicc)*FRACUNIT, P_RandomRange(0, tall)*FRACUNIT, MT_DSNO)
				flake.fuse = 15
				flake.randspeed = P_RandomRange(-10, 10)
				flake.randrate = P_RandomRange(-20, 20)
				P_InstaThrust(flake, p.drawangle + ANGLE_180, 5*FRACUNIT)
			end

			//Wind
			if not (leveltime%2)
				for i = 1, 2 do --start for

				local wind = P_SpawnMobjFromMobj(p.mo, 0, 0, P_RandomRange(0, 70)*FRACUNIT, MT_DWIN)
				local inv = -1
				if i == 2 then inv = 1 end
				wind.scale = FRACUNIT + FRACUNIT/3
				wind.dad = p.mo
				wind.fuse = 14
				wind.angle = p.mo.angle
				P_InstaThrust(wind, wind.angle, p.speed / 3)
				P_TeleportMove(wind, 
					p.mo.x + 25 * cos(wind.angle + ANGLE_90 * inv),
					p.mo.y + 25 * sin(wind.angle + ANGLE_90 * inv),
					wind.z
				)
				wind.angle = $ - ANG30 * inv

				end 			--end for
			end
		end

	end

	//Deleting our dash shield when it is no longer needed (speed too low or exited dash states)
	if p.dfx.dashShield
		if p.dfx.dashShield.valid and p.dfx.dashShield.scale == FRACUNIT and not p.dfx.dashShield.offsetdone
			p.dfx.dashShield.offsetdone = true
		end
		if ( p.mo.state == S_PLAY_DIRDASH and p.speed < p.normalspeed )
		or ( p.mo.state ~= S_PLAY_DIRDASH and p.mo.state ~= S_PLAY_DIRJUDASH )
			if p.dfx.dashShield.valid
				p.dfx.dashShield.destscale = 1
			end
			p.dfx.dashShield = nil
		end
	end
	
	//Now we'll use the RollReturn to gradually go back to our standing position
	if p.dfx.rollReturn
		local intensity = 0
		if p.dfx.rollReturn > 1 and p.dfx.rollReturn < 4 then intensity = ANG10
		elseif p.dfx.rollReturn > 4 then intensity = ANG15 end
		p.mo.rollangle = intensity*p.dfx.LR
		p.dfx.rollReturn = $-1
	end

	if p.mo.state == S_PLAY_DIRJUDASH and (p.pflags & PF_SHIELDABILITY) then p.mo.state = S_PLAY_JUMP end

	//We did a dash jump, so let's create some afterimages
	if p.mo.state == S_PLAY_DIRJUDASH
		if not (leveltime%3) then P_SpawnGhostMobj(p.mo) end
	end

	//We aren't dashing nor jump dashing, so check some stuff
	if p.mo.state ~= S_PLAY_DIRDASH and p.mo.state ~= S_PLAY_DIRJUDASH

		//Turn off jesus powers
		p.charflags = $ & ~SF_RUNONWATER

		//Correct possible liftoff goofs
		p.dfx.dashLiftoff = false

		//Stamina stuff here
		if p.dfx.stamina < MAXSTAMINA
			if p.panim == PA_ROLL or p.panim == PA_JUMP or p.mo.state == S_PLAY_DIRSPINSTOMP// or p.dfx.stasisTime
				if not (leveltime%4) then p.dfx.stamina = $+1 end
			else
				p.dfx.stamina = $+1
			end
		end
	end
	if p.dfx.stamina <= 0 then p.dfx.noStam = true end
	if p.dfx.stamina >= MAXSTAMINA-5 then p.dfx.noStam = false end

	//Our stomp ended, so let's correct any possible squash n stretch oddities
	if p.dfx.correctSquash and p.mo.state ~= S_PLAY_DIRSPINSTOMP
		p.mo.spritexscale = p.dfx.squashxscale
		p.mo.spriteyscale = p.dfx.squashyscale
		p.dfx.correctSquash = nil
	end

	//A spring launched us during the stomp state, so let's get a boost
	if p.mo.eflags & MFE_SPRUNG and p.dfx.lastState == S_PLAY_DIRSPINSTOMP
		P_SetObjectMomZ(p.mo, p.mo.momz/6 + FRACUNIT, true)
		S_StartSound(p.mo, sfx_sprong)
		S_StopSoundByID(p.mo, sfx_spring)
	end

	//We already spawned one ice spike ring, so now let's spawn the rest
	if p.dfx.spikesToSpawn
		p.dfx.spikeTimer = $-1
		if p.dfx.spikeTimer <= 0
			p.dfx.spikesSpawned = $+1
			spikeSpawner(
				p.mo, 
				DIRSPIKEPERCIRCLE, 
				FRACUNIT + ((FRACUNIT/8) * p.dfx.spikesSpawned),
				p.dfx.spikean,
				SPIKEDIST + (SPIKEDIST/3) * p.dfx.spikesSpawned, 
				p.dfx.spikex, 
				p.dfx.spikey, 
				p.dfx.spikez
			)
			if p.dfx.spikesSpawned >= p.dfx.spikesToSpawn
				p.dfx.spikesToSpawn = 0
				p.dfx.spikesSpawned = 0
				p.dfx.spikeTimer = 0
				p.dfx.spikean = nil
				p.dfx.spikex = 0
				p.dfx.spikey = 0
				p.dfx.spikez = 0
			else
				p.dfx.spikeTimer = SPIKECOOLDOWN
			end
		end
	end

	//Keeping our last state around so we can check for landing off a dash jump and some other stuff
	p.dfx.lastState = p.mo.state

	//Our previous mobj angle, for rollangle tweaks
	p.dfx.lastAngle = p.mo.angle

end)

//PostThink for dealing with coming back to the ground during a dash
addHook("PostThinkFrame", function()
	for p in players.iterate
		if not validDirf(p.mo) continue end

		//We landed from a jump dash. Return to dash mode unless we are holding spin.
		if ((p.dfx.lastState == S_PLAY_DIRJUDASH or p.dfx.lastState == S_PLAY_JUMP or p.dfx.lastState == S_PLAY_FALL or p.dfx.dashLiftoff)
		or (p.dfx.lastState == S_PLAY_ROLL and p.mo.state ~= p.dfx.lastState))
		and (p.cmd.buttons & BT_CUSTOM1)
		and P_IsObjectOnGround(p.mo)
		and (p.dfx.stamina > 0 and not p.dfx.noStam)
			p.mo.state = S_PLAY_DIRDASH
			dirPrepDash(p)
			p.dfx.dashLiftoff = false
		end

		//Repeated code for grounded springs, sue me. I don't wanna add any more bull to the above AND OR chain
		if (p.mo.eflags & MFE_SPRUNG) and P_IsObjectOnGround(p.mo)
		and (p.dfx.lastState == S_PLAY_RUN or p.dfx.lastState == S_PLAY_WALK or p.dfx.lastState == S_PLAY_DIRDASH)
		and (p.dfx.stamina > 0 and not p.dfx.noStam) and (p.cmd.buttons & BT_CUSTOM1)
			p.mo.state = S_PLAY_DIRDASH
			dirPrepDash(p)
			p.dfx.dashLiftoff = false
			if not p.powers[pw_sneakers] then p.powers[pw_sneakers] = 35 end
		end

		//We are holding spin, so land with a spin
		if p.dfx.lastState == S_PLAY_DIRJUDASH and P_IsObjectOnGround(p.mo) and p.cmd.buttons & BT_SPIN 
			p.mo.state = S_PLAY_ROLL
			S_StartSound(p.mo, sfx_spin)
		end

		//Spin stomp enemy search and damaging
		if p.dfx.lastState == S_PLAY_DIRSPINSTOMP and P_IsObjectOnGround(p.mo)
			//Did we land on something bouncy? If so, forget the stomp and launch yourself up.
			if P_PlayerTouchingSectorSpecial(p, 1, 15)
				P_DoJump(p, false)
				P_SetObjectMomZ(p.mo, 20*FRACUNIT)
				p.mo.state = S_PLAY_JUMP
			else
			//Otherwise do a blockmap search to find enemies to toss around
				S_StartSound(p.mo, sfx_s3k9b)
				local searchRange = 150*FRACUNIT

				//Blockmap search function
				searchBlockmap("objects", function(pmo, mo)

					if ( (mo.flags & MF_ENEMY or mo.flags & MF_BOSS or mo.flags & MF_MONITOR or mo.type == MT_SPIKE)
					or   (mo.player and mo.player.valid and ( (gametyperules & GTR_RINGSLINGER) or CV_FindVar("friendlyfire").value) ) )
					and P_IsObjectOnGround(mo)
					and abs(pmo.z - mo.z) < 40*FRACUNIT
					and R_PointToDist2(pmo.x, pmo.y, mo.x, mo.y) < STOMPRANGE*FRACUNIT

						// Egg guard
						if mo.type == MT_EGGGUARD and mo.tracer 
							P_KillMobj(mo.tracer, pmo, pmo, DMG_NUKE)
						end

						// Spikes
						if mo.type == MT_SPIKE and mo.valid then P_KillMobj(mo, pmo, pmo, DMG_NUKE) return end

						// Players
						if mo.player
							if (mo.player.powers[pw_super] or mo.player.powers[pw_invulnerability]
							or mo.player.powers[pw_flashing] or (mo.player.pflags & PF_GODMODE)) return end

							if ( (gametyperules & GTR_TEAMS) and pmo.player.ctfteam == mo.player.ctfteam and CV_FindVar("friendlyfire").value )
							or ( (gametyperules & GTR_FRIENDLY) and CV_FindVar("friendlyfire").value )
								S_StartSound(pmo, sfx_s3k5d)
								P_DoPlayerPain(mo.player, pmo, pmo)
								return

							elseif ( (gametyperules & GTR_TEAMS) and pmo.player.ctfteam ~= mo.player.ctfteam )
							or ( not (gametyperules & GTR_TEAMS) and not (gametyperules & GTR_FRIENDLY) )
								P_DamageMobj(mo, pmo, pmo)
								if pmo.player.dfx then pmo.player.dfx.nofling = 3 end
								return
							else return end
						end

						// Everything else
						P_DamageMobj(mo, pmo)
					end

				end, p.mo,
				p.mo.x - searchRange, p.mo.x + searchRange, p.mo.y - searchRange, p.mo.y + searchRange)

				//Floor busting
				for rover in p.mo.subsector.sector.ffloors()
					//Sanity check for bustable FOFs
					if not (rover.flags & FF_EXISTS) or not (rover.flags & FF_BUSTUP) then continue end
					if p.mo.z >= rover.topheight + p.mo.height 
					or p.mo.z + p.mo.momz + p.mo.height < rover.bottomheight
						continue
					end
					EV_CrumbleChain(sector, rover)
					p.mo.state = S_PLAY_JUMP
				end

				//If we're super then just do a big ol nuke blast
				if p.powers[pw_super]
					P_NukeEnemies(p.mo, p.mo, (STOMPRANGE + STOMPRANGE/2)*FRACUNIT)
					S_StartSound(p.mo, sfx_kc40)
				else
				//Otherwise create a shockwave ring
					local stompring = P_SpawnMobjFromMobj(p.mo, 0, 0, 14*FRACUNIT * P_MobjFlip(p.mo), MT_DSTOMPRING)
					stompring.scale = 1
					stompring.flags2 = $|MF2_SPLAT
					stompring.destscale = FRACUNIT*5
					stompring.scalespeed = FRACUNIT/3
					p.mo.momx = 0
					p.mo.momy = 0
				end

				//And then we spawn some ice spikes using a function
				//We'll also set a timer to spawn more spikes shortly after the first one, adding even more if we're super
				spikeSpawner(p.mo, DIRSPIKEPERCIRCLE, FRACUNIT + FRACUNIT/8, p.mo.angle, SPIKEDIST, p.mo.x, p.mo.y, p.mo.z)
				p.dfx.spikesToSpawn = DIRSPIKECIRCLES
				if p.powers[pw_super] then p.dfx.spikesToSpawn = $ + 2 end
				p.dfx.spikeTimer = SPIKECOOLDOWN
				p.dfx.spikesSpawned = 1
				p.dfx.spikean = p.mo.angle
				p.dfx.spikex = p.mo.x
				p.dfx.spikey = p.mo.y
				p.dfx.spikez = p.mo.z

				//Finally, boost us back upwards a tiny bit (or a lot if we're super)
				P_DoJump(p, false)
				if not p.powers[pw_super]
					P_SetObjectMomZ(p.mo, (p.mo.eflags & MFE_UNDERWATER and 7 or 12) * FRACUNIT)
				else
					P_SetObjectMomZ(p.mo, (p.mo.eflags & MFE_UNDERWATER and 14 or 20) * FRACUNIT)
				end
				p.mo.state = S_PLAY_JUMP
				p.pflags = $|PF_THOKKED
				if p == consoleplayer then P_StartQuake(p.powers[pw_super] and 25*FRACUNIT or 15*FRACUNIT, 5) end
			end
		end

		if p.mo.state == S_DRIP and p.dorflex
			p.mo.sprite = SPR_DORF
			p.mo.spritexscale = FRACUNIT/2
			p.mo.spriteyscale = FRACUNIT/2
			p.mo.shadowscale = 1
		end

		if p.dfx.lastState == S_DRIP and p.mo.state ~= S_DRIP
			p.mo.spritexscale = FRACUNIT
			p.mo.spriteyscale = FRACUNIT
			p.mo.shadowscale = FRACUNIT
		end

	end
end)

//We jumped. Let's check if we were dashing and, if so, start the jump dash variant
addHook("JumpSpecial", function(p)
	if not validDirf(p.mo) return end
	
	//Let's set our ice bridge status to null and disable its creation ability
	//That way we can make a new "leading piece" if we land on water from a dash jump.
	if p.dfx.bridge and p.dfx.bridge.valid
		p.dfx.bridgejump = p.dfx.bridge
		p.dfx.bridge.first = false
		p.dfx.bridge = nil
	end

	//We'll make ourselves jump a little lower and without turning into a spin ball, as long as playing a different sound
	if p.mo.state == S_PLAY_DIRDASH and not (p.pflags & PF_JUMPDOWN) and p.speed > p.normalspeed - 3*FRACUNIT
	and (P_IsObjectOnGround(p.mo) or p.dashcoyote)
		if not (p.powers[pw_sneakers] or p.powers[pw_super]) then p.dfx.stamina = $ - STAMDAMAGE_JUMP end
		if p.dfx.stamina < 0 then p.dfx.stamina = 0 end
		p.dfx.dashLiftoff = false
		P_DoJump(p, false)
		p.pflags = $|PF_NOJUMPDAMAGE
		p.mo.momz = $ - ($/4)
		S_StartSound(p.mo, sfx_s23c)
		p.mo.state = S_PLAY_DIRJUDASH
		//If we aren't close to our top speed just jump normally, otherwise give yourself a small speed boost!
		if p.dfx.speed and p.speed > p.normalspeed then P_InstaThrust(p.mo, p.mo.angle, p.dfx.speed + 8*FRACUNIT) end
		return true
	end
end)

//For recurling during an air dash
addHook("AbilitySpecial", function(p)
	if not validDirf(p.mo) return end

	if p.mo.state == S_PLAY_DIRJUDASH and not (p.pflags & PF_JUMPDOWN)
		p.mo.state = S_PLAY_JUMP
		p.pflags = $|PF_JUMPED & ~PF_SPINNING & ~PF_NOJUMPDAMAGE
		S_StartSound(p.mo, sfx_spin)
		return
	end
end)

//Let's do a jump spin stomp
addHook("JumpSpinSpecial", function(p)
	if not validDirf(p.mo) return end

	if not (p.pflags & PF_SPINDOWN) and p.powers[pw_shield] and p.mo.state == S_PLAY_DIRJUDASH and not (p.pflags & PF_SHIELDABILITY)
	and not (p.powers[pw_super])
		p.mo.state = S_PLAY_JUMP
		return
	end

	if not (p.pflags & PF_SPINDOWN) and not (p.pflags & PF_THOKKED) 
	and (not (p.powers[pw_shield]) or (p.powers[pw_super]) )
	and not (p.mo.eflags & MFE_GOOWATER)
		prepSpinStomp(p)
	end
end)

//Let's smash up some ground spikes, but we'll leave the wall ones alone
addHook("MobjCollide", function(mo, pmo)
	if not validDirf(pmo) return end

	if (pmo.z > mo.z + mo.height + FRACUNIT*10) or (mo.z > pmo.z + pmo.height) return end
	if (pmo.state == S_PLAY_DIRDASH or pmo.state == S_PLAY_DIRJUDASH) and pmo.player.speed >= pmo.player.normalspeed
		S_StartSound(pmo, mo.info.deathsound)
		mo.health = 0
		P_KillMobj(mo, pmo, pmo)
		return false
	end
end, MT_SPIKE)

//And also bust up some walls
addHook("MobjLineCollide", function(mo, line)
	if not validDirf(mo) return end

	if (mo.state == S_PLAY_DIRDASH or mo.state == S_PLAY_DIRJUDASH) and mo.player.speed >= mo.player.normalspeed
		for _, sec in ipairs({line.frontsector, line.backsector})
			for rover in sec.ffloors()
				if not (rover.flags & FF_EXISTS) or not (rover.flags & FF_BUSTUP) then continue end
				if mo.z >= rover.topheight
				or mo.z + mo.height < rover.bottomheight
					continue
				end
				EV_CrumbleChain(sector, rover)
			end
		end
	end
end, MT_PLAYER)

//Checking for damage from touching lava specifically
addHook("MobjDamage", function(pmo, inflictor, source, dmgtype)
	if not validDirf(pmo) return end

	if ((pmo.eflags & MFE_TOUCHLAVA) 
	or (inflictor and inflictor.type == MT_LAVAFALL_LAVA))
	and not pmo.player.dfx.icebarrier
		S_StartSound(pmo, sfx_shattr)
		S_StartSound(pmo, sfx_shldls)
		pmo.player.rings = 0
		P_KillMobj(pmo, nil, nil, DMG_FIRE)
		return true
	end

	if not pmo.player.powers[pw_shield] and pmo.player.dfx.icebarrier
		S_StartSound(pmo, sfx_shattr)
		S_StartSound(pmo, sfx_shldls)
		P_DoPlayerPain(pmo.player, source, inflictor)
		pmo.player.dfx.icebarrier = false
		if pmo.player.dfx.iceshield
			for i = 1, #pmo.player.dfx.iceshield
				if pmo.player.dfx.iceshield[i].valid then pmo.player.dfx.iceshield[i].dad = nil end
			end
			pmo.player.dfx.iceshield = nil
		end
		pmo.player.dfx.barriercooldown = BARRIERTICS * TICRATE + TICRATE/2
		local shards = P_RandomRange(8, 10)
		for i = 1, shards do
			local ang = i * FixedAngle( (360/shards)*FRACUNIT )
			local ice = P_SpawnMobj(pmo.x, pmo.y, pmo.z + (pmo.height/4)*3, MT_ICESHATTER)
			ice.angle = pmo.angle
			ice.scale = $ + FixedMul((FRACUNIT/4) * P_RandomRange(-2, 2), pmo.scale)
			ice.fuse = 70
			P_InstaThrust(ice, ice.angle + i*ANG60, P_RandomRange(3, 5) * FRACUNIT)
			P_SetObjectMomZ(ice, P_RandomRange(5, 2) * FRACUNIT)
		end
		return true
	end

end, MT_PLAYER)

//We bumped into something. Check if it's an enemy and whack 'em
addHook("TouchSpecial", thingYeeter)

//Replenish extra oxygen when touching a bubble
addHook("MobjDeath", function(mo, pmo, pmos)
	if not validDirf(pmo) and not validDirf(pmos) return end

	if pmo and validDirf(pmo) then pmo.player.powers[pw_underwater] = 60*TICRATE
	elseif pmos and validDirf(pmos) then pmos.player.powers[pw_underwater] = 60*TICRATE end
end, MT_EXTRALARGEBUBBLE)

//Allow dirfex to damage things in his spin stomp state
addHook("PlayerCanDamage", function(p, mo)
	if not p.valid or not p.mo.valid or not mo.valid return end
	if p.mo.state == S_PLAY_DIRSPINSTOMP return true end
end)

//Crashing against walls and shoving monitors outta the way
addHook("MobjMoveBlocked", dashBonk, MT_PLAYER)

addHook("MobjCollide", function(collider, pmo)
	if not validDirf(pmo) return end
	if not collider.valid or not collider.player or not collider.player.valid return end
	if not CV_FindVar("friendlyfire").value and not (gametyperules & GTR_RINGSLINGER) return end
	if abs(collider.z - pmo.z) > collider.height return end
	
	if pmo.state == S_PLAY_DIRDASH and
	(not collider.player.powers[pw_super] and not collider.player.powers[pw_invulnerability]
     and not collider.player.powers[pw_flashing] and not (collider.player.pflags & PF_GODMODE))

		if ( (gametyperules & GTR_TEAMS) and pmo.player.ctfteam == collider.player.ctfteam and CV_FindVar("friendlyfire").value )
		or ( (gametyperules & GTR_FRIENDLY) and CV_FindVar("friendlyfire").value )
			S_StartSound(pmo, sfx_s3k5d)
			P_DoPlayerPain(collider.player, pmo, pmo)

		elseif ( (gametyperules & GTR_TEAMS) and pmo.player.ctfteam ~= collider.player.ctfteam )
		or ( not (gametyperules & GTR_TEAMS) and not (gametyperules & GTR_FRIENDLY) )
			P_DamageMobj(collider, pmo, pmo)
			if pmo.player.dfx then pmo.player.dfx.nofling = 3 end
		end
	end
end, MT_PLAYER)

COM_AddCommand("iam", function(p, arg)
	if not p or not p.valid return end
	if not arg S_StartSound(nil, sfx_madru, p) return end
	if arg:lower() == "dorflex"
		S_StartSound(nil, sfx_duin, p)
		if not p.dorflex
			CONS_Printf(p, "\x82The power of forbidden brazilian memes wells up within you.")
			p.dorflex = true
		else
			CONS_Printf(p, "\x82The memes' grasp fades.")
			p.dorflex = false
		end
	else
		S_StartSound(nil, sfx_madru, p)
	end
end)
