
//Variables used elsewhere in the code

local SHIELDIST = 25	  						//How far away dirfex's dash shield sits from him
local BRIDGEGROWSPEED = FRACUNIT/8 				//How fast the ice bridge's parts should grow into their full scale
local ROTSPEED = FixedAngle(2*FRACUNIT)			//How fast the ice shield shiould rotate

//An enemy got YEETED so it's gonna get spun around then take damage as soon as its movement is blocked by something
addHook("MobjThinker", function(mo)
	if not mo.valid return end
	//We're currently being tossed around and haven't hit the ground yet. Spin a bit!
	if mo.killer 
		mo.rollangle = $ + mo.knockAng/8
		mo.flags = $ & ~MF_SPECIAL
		if mo.colortimer 
			mo.colortimer = $-1
			if not mo.colortimer
				mo.color = mo.origcolor
				mo.colorized = false
			end
		end
		if not (leveltime%3)
			local dust = P_SpawnMobjFromMobj(mo, 0, 0, (mo.height/4)*3, MT_SMOKE)
			dust.scale = (FRACUNIT/4)*3
			P_SetObjectMomZ(dust, P_RandomRange(10, 18)*FRACUNIT)
		end
	end
	//We hit the ground, finally. Time to die
	if mo.valid and mo.ohgodohfuck and P_IsObjectOnGround(mo) then
		P_DamageMobj(mo, mo.killer, mo.killer)
		if not mo.valid return end
		mo.ohgodohfuck = false
		mo.killer = nil
		mo.rollangle = 0
		mo.colorized = false
	end
end)

//Thinker for ice shield during dash
addHook("MobjThinker", function(mo)
	//You'll see these pretty often so I won't comment all of them. I should REALLY make a function for these too.
	if not mo.valid return end
	if not mo.dad or not mo.dad.valid P_RemoveMobj(mo) return end
	if not mo.dad.player or not mo.dad.player.valid P_RemoveMobj(mo) return end
	if not mo.dad.player.dfx P_RemoveMobj(mo) return end

	//Remove if tiny
	if mo.scale <= FRACUNIT/20 P_RemoveMobj(mo) return end

	//Make our paper sprite always face "dad", our player
	mo.angle = mo.dad.player.drawangle + ANGLE_90

	//Flip the shield around in reverse gravity
	if mo.dad.eflags & MFE_VERTICALFLIP then mo.eflags = $|MFE_VERTICALFLIP
	else mo.eflags = $ & ~MFE_VERTICALFLIP end

	//In first person mode make our shield invisible
	if mo.dad.player == consoleplayer
		if not camera.chase then mo.flags2 = $|MF2_DONTDRAW
		else mo.flags2 = $ & ~MF2_DONTDRAW end
	end

	//And finally warp to the correct position
	P_TeleportMove(mo, 
                   mo.dad.x + SHIELDIST * cos(mo.dad.player.drawangle),
				   mo.dad.y + SHIELDIST * sin(mo.dad.player.drawangle),
				   mo.dad.z
	)
end, MT_DSHI)

//Thinker for dash snowflake effect
addHook("MobjThinker", function(mo)
	if not mo.valid return end
	if mo.scale <= FRACUNIT/20 P_RemoveMobj(mo) return end

	//Move the snowflake in a sine wave pattern on Z
	if not mo.spawnz then mo.spawnz = mo.z end
	mo.z = mo.spawnz + mo.randspeed * sin( FixedAngle((leveltime%360)*FRACUNIT)*mo.randrate )
	
end, MT_DSNO)

//Thinker for ice bridge parts
addHook("MobjThinker", function(mo)
	if not mo.valid return end
	if mo.scale <= FRACUNIT/20 P_RemoveMobj(mo) return end
	if mo.watertop == mo.z - 1000*FRACUNIT P_RemoveMobj(mo) return end

	//Grow into our proper size. I'm not using mo.destscale here since that changes the bridge's collision area
	if mo.spritexscale < FRACUNIT
		mo.spritexscale = $ + BRIDGEGROWSPEED
		mo.spriteyscale = $ + BRIDGEGROWSPEED
		if mo.spritexscale > FRACUNIT
			mo.spritexscale = FRACUNIT
			mo.spriteyscale = FRACUNIT
		end
	end

	//Dirfex is close to the first bridge's edge. Make a new "leading piece" and make the current one no longer lead anything.
	if mo.dad and mo.dad.valid and mo.dad.player and mo.dad.player.valid and mo.dad.player.dfx

		//We're about to be removed from our fuse instead of the bridge limit, so decrese bridgecount manually
		if mo.fuse == 1 and not mo.removed then mo.dad.player.dfx.bridgecount = $-1 end

		//Let's make a new bridge segment
		if mo.first and CanMakeBridge(mo.dad)
		and R_PointToDist2(mo.x, mo.y, mo.dad.x, mo.dad.y) > mo.radius + mo.radius/3
			local bridge = P_SpawnMobjFromMobj(mo.dad, 0, 0, -6*FRACUNIT, MT_DBRIPART)
			if not (leveltime%2) then bridge.z = $ - (5*FRACUNIT) * P_MobjFlip(mo.dad) end
			bridge.fuse = 5*TICRATE
			bridge.scale = FRACUNIT + FRACUNIT/2
			bridge.spritexscale = 1
			bridge.spriteyscale = 1
			bridge.renderflags = $|RF_NOSPLATBILLBOARD
			bridge.angle = mo.dad.player.drawangle
			bridge.dad = mo.dad
			bridge.first = true
			mo.afterme = bridge
			mo.dad.player.dfx.bridge = bridge
			mo.dad.player.dfx.bridgecount = $+1
			if not (leveltime%5)
				S_StopSoundByID(mo.dad, sfx_dbri)
				S_StartSound(mo.dad, sfx_dbri)
			end
			mo.first = false
		end
	end

end, MT_DBRIPART)

//Thinker for ice spikes growing and shrinking
addHook("MobjThinker", function(mo)
	if not mo.valid return end

	//Make our spike grow into its proper Y size and then, when its fuse is zero, shrink into nothingness
	if mo.spriteyscale ~= mo.destyscale and mo.fuse
		mo.spriteyscale = $ + FRACUNIT/3
		if mo.spriteyscale > mo.destyscale then mo.spriteyscale = mo.destyscale end
	elseif mo.fuse == 0
		mo.spriteyscale = $ - FRACUNIT/5
		mo.frame = $+FF_TRANS10
		if mo.spriteyscale < 200 then P_RemoveMobj(mo) return end
	end

end, MT_DICESPIKE)

//Make our ice spikes damage enemies, but never players
addHook("MobjMoveCollide", function(spike, mo)
	if not spike.valid or not mo.valid return end
	if mo.player return end
	if not (mo.flags & MF_ENEMY) and not (mo.flags & MF_MONITOR) and not (mo.flags & MF_BOSS) return end
	if mo.icedamaged and mo.icedamaged == spike then return end

	//Damage 'em if the spike's height is just right
	if abs(mo.z - spike.z) < spike.height
		mo.icedamaged = spike
		P_DamageMobj(mo, spike, spike.target, 1, DMG_SPIKE)
	end
end, MT_DICESPIKE)

//Thinker for stomp ring
addHook("MobjThinker", function(mo)
	if not mo.valid return end
	mo.frame = $+FF_TRANS10 //fade the stomp ring out over the course of 10 tics
end, MT_DSTOMPRING)

//Spinning ice shards
addHook("MobjThinker", function(mo)
	if not mo.valid return end

	//Make our ice shards spin around in the air but only do it very slowly on the ground
	if not mo.rollan then mo.rollan = FixedAngle(P_RandomRange(4, 8) * FRACUNIT) end
	if not P_IsObjectOnGround(mo) 
		mo.rollangle = $ + mo.rollan
	else
		mo.rollangle = $ + mo.rollan/8
	end
end, MT_ICESHATTER)

//Orbiting ice barrier papersprites
addHook("MobjThinker", function(mo)
	if not mo.valid return end
	if not mo.dad or not mo.dad.valid or not mo.dad.player or not mo.dad.player.valid P_RemoveMobj(mo) return end
	if not mo.dad.player.dfx P_RemoveMobj(mo) return end

	//In first person mode make our shield invisible
	if mo.dad.player == consoleplayer
		if not camera.chase then mo.flags2 = $|MF2_DONTDRAW
		else mo.flags2 = $ & ~MF2_DONTDRAW end
	end

	mo.destscale = mo.dad.scale/3

	//Slowly rotate our ice barrier at all times.
	mo.angle = $ + ROTSPEED
	P_TeleportMove(
		mo, 
		mo.dad.x + FixedMul(25 * cos(mo.angle + ANGLE_90), mo.dad.scale),
		mo.dad.y + FixedMul(25 * sin(mo.angle + ANGLE_90), mo.dad.scale),
		mo.dad.z + ((mo.dad.height/5)*2) + mo.updown * sin( FixedAngle(leveltime*FRACUNIT) ) * 3
	)
end, MT_DICESHIELD)

//Wind particles
addHook("MobjThinker", function(mo)
	if not mo.valid return end

	if mo.fuse < 9
		mo.spriteyscale = $ - FRACUNIT/5
		mo.frame = $ + FF_TRANS20
	end

end, MT_DWIN)

addHook("TouchSpecial", function(mo, pmo)
	if not mo.valid return end
	if not pmo.valid or not pmo.player or not pmo.player.valid return end

	if pmo.skin == "dirfex" and pmo.player.dfx and pmo.player.dfx.nofling return true end
end, MT_FLINGRING)

//Generic "shrink on fuse end" function
local function FuseShrink(mo)
	mo.destscale = 1 return true
end

//Fuse shrink hooks
addHook("MobjFuse", FuseShrink, MT_DSNO)
addHook("MobjFuse", FuseShrink, MT_DBRIPART)
addHook("MobjFuse", FuseShrink, MT_ICESHATTER)

//Ignore the fuse's end result, we'll use it for something else in our thinker
addHook("MobjFuse", function(mo) return true end, MT_DICESPIKE)

//Killing a yeeted enemy when their movement is blocked
addHook("MobjMoveBlocked", function(mo, thing, line)
	if mo.valid and mo.ohgodohfuck then 
		P_DamageMobj(mo, mo.killer, mo.killer)
		if not mo.valid return end
		mo.ohgodohfuck = false
		mo.killer = nil
		mo.rollangle = 0
		mo.colorized = false
	end
end)